/*
	GP3200 EMU maincode
	cording by taka
*/

#include "gp3200.h"
#include <stdlib.h>

typedef void(*gp3200_operator)(gp3200_cpu *cpu);

void gp3200_nop(gp3200_cpu *cpu);
void gp3200_mov(gp3200_cpu *cpu);
void gp3200_add(gp3200_cpu *cpu);
void gp3200_sub(gp3200_cpu *cpu);
void gp3200_imul(gp3200_cpu *cpu);
void gp3200_idiv(gp3200_cpu *cpu);
void gp3200_and(gp3200_cpu *cpu);
void gp3200_or(gp3200_cpu *cpu);
void gp3200_not(gp3200_cpu *cpu);
void gp3200_xor(gp3200_cpu *cpu);
void gp3200_int(gp3200_cpu *cpu);
void gp3200_call(gp3200_cpu *cpu);
void gp3200_ret(gp3200_cpu *cpu);
void gp3200_jmp(gp3200_cpu *cpu);
void gp3200_cmp(gp3200_cpu *cpu);
void gp3200_sar(gp3200_cpu *cpu);
void gp3200_sal(gp3200_cpu *cpu);
void gp3200_je(gp3200_cpu *cpu);
void gp3200_jae(gp3200_cpu *cpu);
void gp3200_jbe(gp3200_cpu *cpu);
void gp3200_ja(gp3200_cpu *cpu);
void gp3200_jb(gp3200_cpu *cpu);
void gp3200_jz(gp3200_cpu *cpu);
void gp3200_jnz(gp3200_cpu *cpu);
void gp3200_jne(gp3200_cpu *cpu);
void gp3200_in(gp3200_cpu *cpu);
void gp3200_out(gp3200_cpu *cpu);
void gp3200_push(gp3200_cpu *cpu);
void gp3200_pop(gp3200_cpu *cpu);
void gp3200_cli(gp3200_cpu *cpu);
void gp3200_sti(gp3200_cpu *cpu);
void gp3200_fmov(gp3200_cpu *cpu);
void gp3200_fadd(gp3200_cpu *cpu);
void gp3200_fsub(gp3200_cpu *cpu);
void gp3200_fmul(gp3200_cpu *cpu);
void gp3200_fdiv(gp3200_cpu *cpu);

gp3200_operator operator_table[] = {
	gp3200_nop,
	gp3200_mov,
	gp3200_add,
	gp3200_sub,
	gp3200_imul,
	gp3200_idiv,
	gp3200_and,
	gp3200_or,
	gp3200_not,
	gp3200_xor,
	gp3200_int,
	gp3200_call,
	gp3200_ret,
	gp3200_jmp,
	gp3200_cmp,
	gp3200_sar,
	gp3200_sal,
	gp3200_je,
	gp3200_jae,
	gp3200_jbe,
	gp3200_ja,
	gp3200_jb,
	gp3200_jz,
	gp3200_jnz,
	gp3200_jne,
	gp3200_in,
	gp3200_out,
	gp3200_push,
	gp3200_pop,
	gp3200_cli,
	gp3200_sti,
	gp3200_fmov,
	gp3200_fadd,
	gp3200_fsub,
	gp3200_fmul,
	gp3200_fdiv
};

void gp3200_get_int1(UINT32 **data, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type;
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);
	
	type = cpu->r_memory[cpu->r_pc + 1] & 0xf;

	cpu->r_pc += 2;
	
	switch(type) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data = &cpu->r_regs[args1 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data = (UINT32 *)&cpu->r_memory[cpu->r_regs[args1 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data = (UINT32 *)&cpu->r_memory[args1 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data = args1;
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data = &cpu->r_segs[args1 & 0xff];
		break;
	}
};
	
void gp3200_get_int2(UINT32 **data1, UINT32 **data2, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type[2];
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);
	UINT32 args2;
	
	type[0] = cpu->r_memory[cpu->r_pc + 1] & 0xf;
	type[1] = (cpu->r_memory[cpu->r_pc + 1] >> 4) & 0xf;

	cpu->r_pc += 2;
	
	switch(type[0]) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data1 = &cpu->r_regs[args1 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data1 = (UINT32 *)&cpu->r_memory[cpu->r_regs[args1 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data1 = (UINT32 *)&cpu->r_memory[args1 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data1 = args1;
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data1 = &cpu->r_segs[args1 & 0xff];
		break;
	}
	
	args2 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 0]);
	
	switch(type[1]) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data2 = &cpu->r_regs[args2 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data2 = (UINT32 *)&cpu->r_memory[cpu->r_regs[args2 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data2 = (UINT32 *)&cpu->r_memory[args2 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data2 = args2;
		break;
	case ARGS_FR:
		cpu->r_pc += 1;
		**data2 = (UINT32)gp3200_f32_getfloat(cpu->f_regs[args2 & 0xff]);
		if(type[0] == ARGS_MR || type[0] == ARGS_ME) *data2 = (UINT32 *)&cpu->f_regs[args2 & 0xff];
		break;
	case ARGS_FE:
		cpu->r_pc += 5;
		**data2 = (UINT32)gp3200_f32_getfloat(gp3200_f32_int2float(args2));
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data2 = &cpu->r_segs[args2 & 0xff];
		break;
	}
};

void gp3200_get_float2(gp3200_float **data1, gp3200_float **data2, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type[2];
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);
	UINT32 args2;
	
	type[0] = cpu->r_memory[cpu->r_pc + 1] & 0xf;
	type[1] = (cpu->r_memory[cpu->r_pc + 1] >> 4) & 0xf;

	cpu->r_pc += 2;
	
	switch(type[0]) {
	case ARGS_FR:
		cpu->r_pc += 1;
		*data1 = &cpu->f_regs[args1 & 0xff];
		break;
	}
	
	args2 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 0]);
	
	switch(type[1]) {
	case ARGS_FR:
		cpu->r_pc += 1;
		*data2 = &cpu->f_regs[args2 & 0xff];
		break;
	case ARGS_IR:
		cpu->r_pc += 1;
		**data2 = gp3200_f32_int2float(cpu->r_regs[args2 & 0xff]);
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		**data2 = gp3200_f32_int2float(cpu->r_memory[cpu->r_regs[args2 & 0xff] + cpu->r_segs[1]]);
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		**data2 = gp3200_f32_int2float(cpu->r_memory[args2 + cpu->r_segs[1]]);
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data2 = gp3200_f32_int2float(args2);
		break;
	case ARGS_FE:
		cpu->r_pc += 5;
		**data2 = *((gp3200_float *)&args2);
		break;
	}
};

void gp3200_init(gp3200_cpu *cpu, UINT8 *memory)
{
	int i;
	cpu->r_port = (UINT32 *)malloc(256 * sizeof(UINT32));
	cpu->r_if = 1;
	for(i = 0; i < 8; i++) cpu->r_segs[i] = 0;
	cpu->r_memory = memory;
	cpu->on_int = 0;
}

void gp3200_step(gp3200_cpu *cpu)
{
	if(cpu->r_memory[(cpu->r_pc & (0x80000000 - 1)) + cpu->r_segs[0]] > (sizeof(operator_table)/sizeof(gp3200_operator))) return;
	operator_table[cpu->r_memory[(cpu->r_pc & (0x80000000 - 1)) + cpu->r_segs[0]]](cpu);
}

void gp3200_nop(gp3200_cpu *cpu)
{
	cpu->r_pc++;
}

void gp3200_mov(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 = *data2;
}

void gp3200_add(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 += *data2;
}

void gp3200_sub(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 -= *data2;
}

void gp3200_imul(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 *= *data2;
}

void gp3200_idiv(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 /= *data2;
}

void gp3200_and(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 &= *data2;
}

void gp3200_or(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 |= *data2;
}

void gp3200_not(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	*data = ~*data;
}

void gp3200_xor(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 ^= *data2;
}

void gp3200_int(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_if == 1) {
		if(cpu->on_int == 0) cpu->r_pc = cpu->r_memory[*data * 4];
		else cpu->on_int(cpu, *data, cpu->r_memory);
	}
}

void gp3200_call(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	cpu->r_ret = cpu->r_pc;
	cpu->r_pc = *data;
}

void gp3200_ret(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	cpu->r_pc = cpu->r_ret;
}

void gp3200_jmp(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	cpu->r_pc = *data;
}

void gp3200_cmp(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	if(*data1 == *data2     ) cpu->r_cmp |= 1 << 0;
	else if(*data1 >= *data2     ) cpu->r_cmp |= 1 << 1;
	else if(*data1 <= *data2     ) cpu->r_cmp |= 1 << 2;
	else if(*data1 >  *data2     ) cpu->r_cmp |= 1 << 3;
	else if(*data1 <  *data2     ) cpu->r_cmp |= 1 << 4;
	else if(*data1 -  *data2 == 0) cpu->r_cmp |= 1 << 5;
	else if(*data1 -  *data2 != 0) cpu->r_cmp |= 1 << 6;
	else if(*data1 != *data2     ) cpu->r_cmp |= 1 << 7;
}

void gp3200_sar(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 >>= *data2;
}

void gp3200_sal(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data1 <<= *data2;
}

void gp3200_je(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 0))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jae(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 1))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jbe(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 2))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_ja(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 3))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jb(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 4))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jz(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 5))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jnz(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 6))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_jne(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	if(cpu->r_cmp & (1 << 7))
	{
		cpu->r_pc = *data;
		cpu->r_cmp = 0;
	}
}

void gp3200_in(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	*data2 = cpu->r_port[*data1];
}

void gp3200_out(gp3200_cpu *cpu)
{
	unsigned int used1, used2;
	unsigned int *data1 = &used1, *data2 = &used2;

	gp3200_get_int2(&data1, &data2, cpu);

	cpu->r_port[*data1] = *data2;
}

void gp3200_push(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	cpu->r_sp /* sp */ -= 4;
	cpu->r_memory[cpu->r_regs[6]] = *data;
}

void gp3200_pop(gp3200_cpu *cpu)
{
	unsigned int used1;
	unsigned int *data = &used1;

	gp3200_get_int1(&data, cpu);

	cpu->r_sp /* sp */ += 4;
	*data = cpu->r_memory[cpu->r_regs[6]];
}

void gp3200_cli(gp3200_cpu *cpu)
{
	cpu->r_pc = 0;
	
	cpu->r_pc += 1;
}

void gp3200_sti(gp3200_cpu *cpu)
{
	cpu->r_pc = 1;
	
	cpu->r_pc += 1;
}

void gp3200_fmov(gp3200_cpu *cpu)
{
	gp3200_float used1, used2;
	gp3200_float *data1 = &used1, *data2 = &used2;

	gp3200_get_float2(&data1, &data2, cpu);

	*data1 = *data2;
}

void gp3200_fadd(gp3200_cpu *cpu)
{
	gp3200_float used1, used2;
	gp3200_float *data1 = &used1, *data2 = &used2;

	gp3200_get_float2(&data1, &data2, cpu);

	*data1 = gp3200_f32_fadd(*data1, *data2);
}

void gp3200_fsub(gp3200_cpu *cpu)
{
	gp3200_float used1, used2;
	gp3200_float *data1 = &used1, *data2 = &used2;

	gp3200_get_float2(&data1, &data2, cpu);

	*data1 = gp3200_f32_fsub(*data1, *data2);
}

void gp3200_fmul(gp3200_cpu *cpu)
{
	gp3200_float used1, used2;
	gp3200_float *data1 = &used1, *data2 = &used2;

	gp3200_get_float2(&data1, &data2, cpu);

	*data1 = gp3200_f32_fmul(*data1, *data2);
}

void gp3200_fdiv(gp3200_cpu *cpu)
{
	gp3200_float used1, used2;
	gp3200_float *data1 = &used1, *data2 = &used2;

	gp3200_get_float2(&data1, &data2, cpu);

	*data1 = gp3200_f32_fdiv(*data1, *data2);
}
